# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from .equipment_character_select import EquipmentCharacterSelect
from ...gender import Gender
from ...admin_level import AdminLevel
from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class CharacterSelectionListEntry:
    """
    Character selection screen character
    """
    _byte_size: int = 0
    _name: str
    _id: int
    _level: int
    _gender: Gender
    _hair_style: int
    _hair_color: int
    _skin: int
    _admin: AdminLevel
    _equipment: EquipmentCharacterSelect

    def __init__(self, *, name: str, id: int, level: int, gender: Gender, hair_style: int, hair_color: int, skin: int, admin: AdminLevel, equipment: EquipmentCharacterSelect):
        """
        Create a new instance of CharacterSelectionListEntry.

        Args:
            name (str): 
            id (int): (Value range is 0-4097152080.)
            level (int): (Value range is 0-252.)
            gender (Gender): 
            hair_style (int): (Value range is 0-252.)
            hair_color (int): (Value range is 0-252.)
            skin (int): (Value range is 0-252.)
            admin (AdminLevel): 
            equipment (EquipmentCharacterSelect): 
        """
        self._name = name
        self._id = id
        self._level = level
        self._gender = gender
        self._hair_style = hair_style
        self._hair_color = hair_color
        self._skin = skin
        self._admin = admin
        self._equipment = equipment

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def name(self) -> str:
        return self._name

    @property
    def id(self) -> int:
        return self._id

    @property
    def level(self) -> int:
        return self._level

    @property
    def gender(self) -> Gender:
        return self._gender

    @property
    def hair_style(self) -> int:
        return self._hair_style

    @property
    def hair_color(self) -> int:
        return self._hair_color

    @property
    def skin(self) -> int:
        return self._skin

    @property
    def admin(self) -> AdminLevel:
        return self._admin

    @property
    def equipment(self) -> EquipmentCharacterSelect:
        return self._equipment

    @staticmethod
    def serialize(writer: EoWriter, data: "CharacterSelectionListEntry") -> None:
        """
        Serializes an instance of `CharacterSelectionListEntry` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (CharacterSelectionListEntry): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.string_sanitization_mode = True
            if data._name is None:
                raise SerializationError("name must be provided.")
            writer.add_string(data._name)
            writer.add_byte(0xFF)
            if data._id is None:
                raise SerializationError("id must be provided.")
            writer.add_int(data._id)
            if data._level is None:
                raise SerializationError("level must be provided.")
            writer.add_char(data._level)
            if data._gender is None:
                raise SerializationError("gender must be provided.")
            writer.add_char(int(data._gender))
            if data._hair_style is None:
                raise SerializationError("hair_style must be provided.")
            writer.add_char(data._hair_style)
            if data._hair_color is None:
                raise SerializationError("hair_color must be provided.")
            writer.add_char(data._hair_color)
            if data._skin is None:
                raise SerializationError("skin must be provided.")
            writer.add_char(data._skin)
            if data._admin is None:
                raise SerializationError("admin must be provided.")
            writer.add_char(int(data._admin))
            if data._equipment is None:
                raise SerializationError("equipment must be provided.")
            EquipmentCharacterSelect.serialize(writer, data._equipment)
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "CharacterSelectionListEntry":
        """
        Deserializes an instance of `CharacterSelectionListEntry` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            CharacterSelectionListEntry: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.chunked_reading_mode = True
            name = reader.get_string()
            reader.next_chunk()
            id = reader.get_int()
            level = reader.get_char()
            gender = Gender(reader.get_char())
            hair_style = reader.get_char()
            hair_color = reader.get_char()
            skin = reader.get_char()
            admin = AdminLevel(reader.get_char())
            equipment = EquipmentCharacterSelect.deserialize(reader)
            reader.chunked_reading_mode = False
            result = CharacterSelectionListEntry(name=name, id=id, level=level, gender=gender, hair_style=hair_style, hair_color=hair_color, skin=skin, admin=admin, equipment=equipment)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"CharacterSelectionListEntry(byte_size={repr(self._byte_size)}, name={repr(self._name)}, id={repr(self._id)}, level={repr(self._level)}, gender={repr(self._gender)}, hair_style={repr(self._hair_style)}, hair_color={repr(self._hair_color)}, skin={repr(self._skin)}, admin={repr(self._admin)}, equipment={repr(self._equipment)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ..version import Version
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class InitInitClientPacket(Packet):
    """
    Connection initialization request.
    This packet is unencrypted.
    """
    _byte_size: int = 0
    _challenge: int
    _version: Version
    _hdid_length: int
    _hdid: str

    def __init__(self, *, challenge: int, version: Version, hdid: str):
        """
        Create a new instance of InitInitClientPacket.

        Args:
            challenge (int): (Value range is 0-16194276.)
            version (Version): 
            hdid (str): (Length must be 252 or less.)
        """
        self._challenge = challenge
        self._version = version
        self._hdid = hdid
        self._hdid_length = len(self._hdid)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def challenge(self) -> int:
        return self._challenge

    @property
    def version(self) -> Version:
        return self._version

    @property
    def hdid(self) -> str:
        return self._hdid

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Init

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Init

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        InitInitClientPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "InitInitClientPacket") -> None:
        """
        Serializes an instance of `InitInitClientPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (InitInitClientPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._challenge is None:
                raise SerializationError("challenge must be provided.")
            writer.add_three(data._challenge)
            if data._version is None:
                raise SerializationError("version must be provided.")
            Version.serialize(writer, data._version)
            writer.add_char(112)
            if data._hdid_length is None:
                raise SerializationError("hdid_length must be provided.")
            writer.add_char(data._hdid_length)
            if data._hdid is None:
                raise SerializationError("hdid must be provided.")
            if len(data._hdid) > 252:
                raise SerializationError(f"Expected length of hdid to be 252 or less, got {len(data._hdid)}.")
            writer.add_fixed_string(data._hdid, data._hdid_length, False)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "InitInitClientPacket":
        """
        Deserializes an instance of `InitInitClientPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            InitInitClientPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            challenge = reader.get_three()
            version = Version.deserialize(reader)
            reader.get_char()
            hdid_length = reader.get_char()
            hdid = reader.get_fixed_string(hdid_length, False)
            result = InitInitClientPacket(challenge=challenge, version=version, hdid=hdid)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"InitInitClientPacket(byte_size={repr(self._byte_size)}, challenge={repr(self._challenge)}, version={repr(self._version)}, hdid={repr(self._hdid)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from typing import Union
from .sit_action import SitAction
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ...coords import Coords
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class ChairRequestClientPacket(Packet):
    """
    Sitting on a chair
    """
    _byte_size: int = 0
    _sit_action: SitAction
    _sit_action_data: 'ChairRequestClientPacket.SitActionData'

    def __init__(self, *, sit_action: SitAction, sit_action_data: 'ChairRequestClientPacket.SitActionData' = None):
        """
        Create a new instance of ChairRequestClientPacket.

        Args:
            sit_action (SitAction): 
            sit_action_data (ChairRequestClientPacket.SitActionData): Data associated with the `sit_action` field.
        """
        self._sit_action = sit_action
        self._sit_action_data = sit_action_data

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def sit_action(self) -> SitAction:
        return self._sit_action

    @property
    def sit_action_data(self) -> 'ChairRequestClientPacket.SitActionData':
        """
        ChairRequestClientPacket.SitActionData: Data associated with the `sit_action` field.
        """
        return self._sit_action_data

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Chair

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Request

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        ChairRequestClientPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "ChairRequestClientPacket") -> None:
        """
        Serializes an instance of `ChairRequestClientPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (ChairRequestClientPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._sit_action is None:
                raise SerializationError("sit_action must be provided.")
            writer.add_char(int(data._sit_action))
            if data._sit_action == SitAction.Sit:
                if not isinstance(data._sit_action_data, ChairRequestClientPacket.SitActionDataSit):
                    raise SerializationError("Expected sit_action_data to be type ChairRequestClientPacket.SitActionDataSit for sit_action " + SitAction(data._sit_action).name + ".")
                ChairRequestClientPacket.SitActionDataSit.serialize(writer, data._sit_action_data)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "ChairRequestClientPacket":
        """
        Deserializes an instance of `ChairRequestClientPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            ChairRequestClientPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            sit_action = SitAction(reader.get_char())
            sit_action_data: ChairRequestClientPacket.SitActionData = None
            if sit_action == SitAction.Sit:
                sit_action_data = ChairRequestClientPacket.SitActionDataSit.deserialize(reader)
            result = ChairRequestClientPacket(sit_action=sit_action, sit_action_data=sit_action_data)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"ChairRequestClientPacket(byte_size={repr(self._byte_size)}, sit_action={repr(self._sit_action)}, sit_action_data={repr(self._sit_action_data)})"

    SitActionData = Union['ChairRequestClientPacket.SitActionDataSit', None]
    """
    Data associated with different values of the `sit_action` field.
    """

    class SitActionDataSit:
        """
        Data associated with sit_action value SitAction.Sit
        """
        _byte_size: int = 0
        _coords: Coords

        def __init__(self, *, coords: Coords):
            """
            Create a new instance of ChairRequestClientPacket.SitActionDataSit.

            Args:
                coords (Coords): 
            """
            self._coords = coords

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def coords(self) -> Coords:
            return self._coords

        @staticmethod
        def serialize(writer: EoWriter, data: "ChairRequestClientPacket.SitActionDataSit") -> None:
            """
            Serializes an instance of `ChairRequestClientPacket.SitActionDataSit` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (ChairRequestClientPacket.SitActionDataSit): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._coords is None:
                    raise SerializationError("coords must be provided.")
                Coords.serialize(writer, data._coords)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "ChairRequestClientPacket.SitActionDataSit":
            """
            Deserializes an instance of `ChairRequestClientPacket.SitActionDataSit` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                ChairRequestClientPacket.SitActionDataSit: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                coords = Coords.deserialize(reader)
                result = ChairRequestClientPacket.SitActionDataSit(coords=coords)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"ChairRequestClientPacket.SitActionDataSit(byte_size={repr(self._byte_size)}, coords={repr(self._coords)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from typing import Union
from collections.abc import Iterable
from .warp_type import WarpType
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class WarpRequestServerPacket(Packet):
    """
    Warp request from server
    """
    _byte_size: int = 0
    _warp_type: WarpType
    _map_id: int
    _warp_type_data: 'WarpRequestServerPacket.WarpTypeData'
    _session_id: int

    def __init__(self, *, warp_type: WarpType, map_id: int, warp_type_data: 'WarpRequestServerPacket.WarpTypeData' = None, session_id: int):
        """
        Create a new instance of WarpRequestServerPacket.

        Args:
            warp_type (WarpType): 
            map_id (int): (Value range is 0-64008.)
            warp_type_data (WarpRequestServerPacket.WarpTypeData): Data associated with the `warp_type` field.
            session_id (int): (Value range is 0-64008.)
        """
        self._warp_type = warp_type
        self._map_id = map_id
        self._warp_type_data = warp_type_data
        self._session_id = session_id

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def warp_type(self) -> WarpType:
        return self._warp_type

    @property
    def map_id(self) -> int:
        return self._map_id

    @property
    def warp_type_data(self) -> 'WarpRequestServerPacket.WarpTypeData':
        """
        WarpRequestServerPacket.WarpTypeData: Data associated with the `warp_type` field.
        """
        return self._warp_type_data

    @property
    def session_id(self) -> int:
        return self._session_id

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Warp

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Request

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        WarpRequestServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "WarpRequestServerPacket") -> None:
        """
        Serializes an instance of `WarpRequestServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (WarpRequestServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._warp_type is None:
                raise SerializationError("warp_type must be provided.")
            writer.add_char(int(data._warp_type))
            if data._map_id is None:
                raise SerializationError("map_id must be provided.")
            writer.add_short(data._map_id)
            if data._warp_type == WarpType.MapSwitch:
                if not isinstance(data._warp_type_data, WarpRequestServerPacket.WarpTypeDataMapSwitch):
                    raise SerializationError("Expected warp_type_data to be type WarpRequestServerPacket.WarpTypeDataMapSwitch for warp_type " + WarpType(data._warp_type).name + ".")
                WarpRequestServerPacket.WarpTypeDataMapSwitch.serialize(writer, data._warp_type_data)
            if data._session_id is None:
                raise SerializationError("session_id must be provided.")
            writer.add_short(data._session_id)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "WarpRequestServerPacket":
        """
        Deserializes an instance of `WarpRequestServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            WarpRequestServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            warp_type = WarpType(reader.get_char())
            map_id = reader.get_short()
            warp_type_data: WarpRequestServerPacket.WarpTypeData = None
            if warp_type == WarpType.MapSwitch:
                warp_type_data = WarpRequestServerPacket.WarpTypeDataMapSwitch.deserialize(reader)
            session_id = reader.get_short()
            result = WarpRequestServerPacket(warp_type=warp_type, map_id=map_id, warp_type_data=warp_type_data, session_id=session_id)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"WarpRequestServerPacket(byte_size={repr(self._byte_size)}, warp_type={repr(self._warp_type)}, map_id={repr(self._map_id)}, warp_type_data={repr(self._warp_type_data)}, session_id={repr(self._session_id)})"

    WarpTypeData = Union['WarpRequestServerPacket.WarpTypeDataMapSwitch', None]
    """
    Data associated with different values of the `warp_type` field.
    """

    class WarpTypeDataMapSwitch:
        """
        Data associated with warp_type value WarpType.MapSwitch
        """
        _byte_size: int = 0
        _map_rid: tuple[int, ...]
        _map_file_size: int

        def __init__(self, *, map_rid: Iterable[int], map_file_size: int):
            """
            Create a new instance of WarpRequestServerPacket.WarpTypeDataMapSwitch.

            Args:
                map_rid (Iterable[int]): (Length must be `2`.) (Element value range is 0-64008.)
                map_file_size (int): (Value range is 0-16194276.)
            """
            self._map_rid = tuple(map_rid)
            self._map_file_size = map_file_size

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def map_rid(self) -> tuple[int, ...]:
            return self._map_rid

        @property
        def map_file_size(self) -> int:
            return self._map_file_size

        @staticmethod
        def serialize(writer: EoWriter, data: "WarpRequestServerPacket.WarpTypeDataMapSwitch") -> None:
            """
            Serializes an instance of `WarpRequestServerPacket.WarpTypeDataMapSwitch` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (WarpRequestServerPacket.WarpTypeDataMapSwitch): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._map_rid is None:
                    raise SerializationError("map_rid must be provided.")
                if len(data._map_rid) != 2:
                    raise SerializationError(f"Expected length of map_rid to be exactly 2, got {len(data._map_rid)}.")
                for i in range(2):
                    writer.add_short(data._map_rid[i])
                if data._map_file_size is None:
                    raise SerializationError("map_file_size must be provided.")
                writer.add_three(data._map_file_size)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "WarpRequestServerPacket.WarpTypeDataMapSwitch":
            """
            Deserializes an instance of `WarpRequestServerPacket.WarpTypeDataMapSwitch` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                WarpRequestServerPacket.WarpTypeDataMapSwitch: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                map_rid = []
                for i in range(2):
                    map_rid.append(reader.get_short())
                map_file_size = reader.get_three()
                result = WarpRequestServerPacket.WarpTypeDataMapSwitch(map_rid=map_rid, map_file_size=map_file_size)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"WarpRequestServerPacket.WarpTypeDataMapSwitch(byte_size={repr(self._byte_size)}, map_rid={repr(self._map_rid)}, map_file_size={repr(self._map_file_size)})"



from enum import EnumMeta


class ProtocolEnumMeta(EnumMeta):
    def __call__(cls, value, names=None, *, module=None, qualname=None, type=None, start=1):
        if names is not None:
            return super().__call__(
                value, names=names, module=module, qualname=qualname, type=type, start=start
            )
        try:
            return super().__call__(value, module=module, qualname=qualname, type=type, start=start)
        except ValueError:
            unrecognized = int.__new__(cls, value)
            unrecognized._name_ = f"Unrecognized({int(value)})"
            unrecognized._value_ = value
            return unrecognized



import xml.etree.ElementTree as ET
from protocol_code_generator.generate.code_block import CodeBlock
from protocol_code_generator.generate.switch_code_generator import SwitchCodeGenerator

from protocol_code_generator.util.xml_utils import (
    get_boolean_attribute,
    get_comment,
    get_int_attribute,
    get_required_string_attribute,
    get_string_attribute,
    get_text,
)


class FieldData:
    def __init__(self, name, type, offset, array):
        self.name = name
        self.type_ = type
        self.offset = offset
        self.array = array


class ObjectGenerationContext:
    def __init__(self):
        self.chunked_reading_enabled = False
        self.reached_optional_field = False
        self.reached_dummy = False
        self.needs_old_writer_length_variable = False
        self.accessible_fields = {}
        self.length_field_is_referenced_map = {}


class ObjectGenerationData:
    def __init__(self, class_name):
        self.class_name = class_name
        self.super_interfaces = []
        self.fields = CodeBlock()
        self.methods = CodeBlock()
        self.init_params = []
        self.init_docstring_params = []
        self.init_body = CodeBlock()
        self.serialize = CodeBlock()
        self.deserialize = CodeBlock()
        self.deserialize_init_arguments = []
        self.auxiliary_types = CodeBlock()
        self.docstring = CodeBlock()
        self.repr_fields = ["byte_size"]

    def add_method(self, method):
        if self.methods:
            self.methods.add_line()
        self.methods.add_code_block(method)

    def add_auxiliary_type(self, type):
        if self.auxiliary_types:
            self.auxiliary_types.add_line()
        self.auxiliary_types.add_code_block(type)


class ObjectCodeGenerator:
    def __init__(self, class_name, type_factory, context=None):
        self._class_name = class_name
        self._type_factory = type_factory
        self._context = ObjectGenerationContext() if context is None else context
        self._data = ObjectGenerationData(class_name)

    def generate_instruction(self, instruction):
        if self._context.reached_dummy:
            raise RuntimeError("<dummy> elements must not be followed by any other elements.")

        instruction_name = instruction.tag

        if instruction_name == "field":
            self._generate_field(instruction)
        elif instruction_name == "array":
            self._generate_array(instruction)
        elif instruction_name == "length":
            self._generate_length(instruction)
        elif instruction_name == "dummy":
            self._generate_dummy(instruction)
        elif instruction_name == "switch":
            self._generate_switch(instruction)
        elif instruction_name == "chunked":
            self._generate_chunked(instruction)
        elif instruction_name == "break":
            self._generate_break()

    @property
    def data(self):
        return self._data

    @property
    def code(self):
        simple_name = self._class_name
        if '.' in simple_name:
            simple_name = simple_name.rsplit('.', 1)[1]

        super_interfaces = (
            f"({', '.join(self._data.super_interfaces)})" if self._data.super_interfaces else ""
        )

        result = (
            CodeBlock()
            .add_line(f"class {simple_name}{super_interfaces}:")
            .indent()
            .add_code_block(self._data.docstring)
            .add_line("_byte_size: int = 0")
            .add_code_block(self._data.fields)
            .add_line()
            .add_code_block(self._generate_init_method())
            .add_line()
            .add_code_block(self._generate_get_byte_size())
            .add_line()
            .add_code_block(self._data.methods)
            .add_line()
            .add_code_block(self._generate_serialize_method())
            .add_line()
            .add_code_block(self._generate_deserialize_method())
            .add_line()
            .add_code_block(self._generate_repr_method())
        )

        if self._data.auxiliary_types:
            result.add_line()
            result.add_code_block(self._data.auxiliary_types)

        result.unindent()

        return result

    def _generate_init_method(self):
        result = CodeBlock().add('def __init__(self')
        if self._data.init_params:
            result.add(', *')
        for param in self._data.init_params:
            result.add(', ')
            result.add_code_block(param)
        result.add_line('):')

        result.indent()
        result.add_line('"""')
        result.add_line(f'Create a new instance of {self._class_name}.')
        if self._data.init_docstring_params:
            result.add_line()
            result.add_line('Args:')
            for docstring_param in self._data.init_docstring_params:
                result.add('    ')
                result.add_code_block(docstring_param)
                result.add_line()
        result.add_line('"""')

        result.add_code_block(self._data.init_body)

        return result

    def _generate_get_byte_size(self):
        return (
            CodeBlock()
            .add_line('@property')
            .add_line('def byte_size(self) -> int:')
            .indent()
            .add_line('"""')
            .add_line('Returns the size of the data that this was deserialized from.')
            .add_line()
            .add_line('Returns:')
            .add_line('    int: The size of the data that this was deserialized from.')
            .add_line('"""')
            .add_line('return self._byte_size')
            .unindent()
        )

    def _generate_serialize_method(self):
        result = (
            CodeBlock()
            .add_line("@staticmethod")
            .add_line(f'def serialize(writer: EoWriter, data: "{self._class_name}") -> None:')
            .indent()
            .add_line('"""')
            .add_line(f'Serializes an instance of `{self._class_name}` to the provided `EoWriter`.')
            .add_line()
            .add_line('Args:')
            .add_line('    writer (EoWriter): The writer that the data will be serialized to.')
            .add_line(f'    data ({self._class_name}): The data to serialize.')
            .add_line('"""')
        )

        if self._context.needs_old_writer_length_variable:
            result.add_line('old_writer_length: int = len(writer)')

        result.add_line('old_string_sanitization_mode: bool = writer.string_sanitization_mode')
        result.begin_control_flow('try')
        result.add_code_block(self._data.serialize)
        result.next_control_flow('finally')
        result.add_line('writer.string_sanitization_mode = old_string_sanitization_mode')
        result.unindent()
        result.unindent()
        result.add_import('EoWriter', 'eolib.data.eo_writer')

        return result

    def _generate_deserialize_method(self):
        constructor = CodeBlock().add(f'{self._class_name}(')
        for i, arg in enumerate(self._data.deserialize_init_arguments):
            if i > 0:
                constructor.add(', ')
            constructor.add_code_block(arg)
        constructor.add_line(')')

        return (
            CodeBlock()
            .add_line("@staticmethod")
            .add_line(f'def deserialize(reader: EoReader) -> "{self._class_name}":')
            .indent()
            .add_line('"""')
            .add_line(
                f'Deserializes an instance of `{self._class_name}` from the provided `EoReader`.'
            )
            .add_line()
            .add_line('Args:')
            .add_line('    reader (EoReader): The writer that the data will be serialized to.')
            .add_line()
            .add_line('Returns:')
            .add_line(f'    {self._class_name}: The data to serialize.')
            .add_line('"""')
            .add_line('old_chunked_reading_mode: bool = reader.chunked_reading_mode')
            .begin_control_flow('try')
            .add_line('reader_start_position: int = reader.position')
            .add_code_block(self._data.deserialize)
            .add('result = ')
            .add_code_block(constructor)
            .add_line('result._byte_size = reader.position - reader_start_position')
            .add_line('return result')
            .next_control_flow('finally')
            .add_line('reader.chunked_reading_mode = old_chunked_reading_mode')
            .unindent()
            .add_import('EoReader', 'eolib.data.eo_reader')
        )

    def _generate_repr_method(self):
        field_to_repr_str = lambda field: field + "={repr(self._" + field + ")}"
        repr_field_strs = [field_to_repr_str(field) for field in self._data.repr_fields]
        repr_expression = 'f"' + self._class_name + '(' + ', '.join(repr_field_strs) + ')"'
        return (
            CodeBlock()
            .add_line('def __repr__(self):')
            .indent()
            .add_line(f'return {repr_expression}')
            .unindent()
        )

    def _generate_field(self, protocol_field):
        optional = protocol_field.get("optional")
        self._check_optional_field(optional)

        field_code_generator = (
            self._field_code_generator_builder()
            .name(get_string_attribute(protocol_field, "name"))
            .type(get_required_string_attribute(protocol_field, "type"))
            .length(get_string_attribute(protocol_field, "length"))
            .padded(get_boolean_attribute(protocol_field, "padded"))
            .optional(optional)
            .hardcoded_value(get_text(protocol_field))
            .comment(get_comment(protocol_field))
            .build()
        )

        field_code_generator.generate_field()
        field_code_generator.generate_serialize()
        field_code_generator.generate_deserialize()

        if optional:
            self._context.reached_optional_field = True

    def _generate_array(self, protocol_array):
        optional = protocol_array.get("optional")
        self._check_optional_field(optional)

        delimited = protocol_array.get("delimited")
        if delimited and not self._context.chunked_reading_enabled:
            raise RuntimeError(
                "Cannot generate a delimited array instruction unless chunked reading is enabled."
            )

        field_code_generator = (
            self._field_code_generator_builder()
            .name(get_required_string_attribute(protocol_array, "name"))
            .type(get_required_string_attribute(protocol_array, "type"))
            .length(get_string_attribute(protocol_array, "length"))
            .optional(optional)
            .comment(get_comment(protocol_array))
            .array_field(True)
            .delimited(delimited)
            .trailing_delimiter(get_boolean_attribute(protocol_array, "trailing-delimiter", True))
            .build()
        )

        field_code_generator.generate_field()
        field_code_generator.generate_serialize()
        field_code_generator.generate_deserialize()

        if optional:
            self._context.reached_optional_field = True

    def _generate_length(self, protocol_length):
        optional = protocol_length.get("optional")
        self._check_optional_field(optional)

        field_code_generator = (
            self._field_code_generator_builder()
            .name(get_required_string_attribute(protocol_length, "name"))
            .type(get_required_string_attribute(protocol_length, "type"))
            .offset(get_int_attribute(protocol_length, "offset"))
            .length_field(True)
            .optional(optional)
            .comment(get_comment(protocol_length))
            .build()
        )

        field_code_generator.generate_field()
        field_code_generator.generate_serialize()
        field_code_generator.generate_deserialize()

        if optional:
            self._context.reached_optional_field = True

    def _generate_dummy(self, protocol_dummy):
        field_code_generator = (
            self._field_code_generator_builder()
            .type(get_required_string_attribute(protocol_dummy, "type"))
            .hardcoded_value(get_text(protocol_dummy))
            .comment(get_comment(protocol_dummy))
            .build()
        )

        needs_if_guards = not self._data.serialize.empty or not self._data.deserialize.empty

        if needs_if_guards:
            self._data.serialize.begin_control_flow("if len(writer) == old_writer_length")
            self._data.deserialize.begin_control_flow("if reader.position == reader_start_position")

        field_code_generator.generate_serialize()
        field_code_generator.generate_deserialize()

        if needs_if_guards:
            self._data.serialize.unindent()
            self._data.deserialize.unindent()

        self._context.reached_dummy = True

        if needs_if_guards:
            self._context.needs_old_writer_length_variable = True

    def _field_code_generator_builder(self):
        from protocol_code_generator.generate.field_code_generator import FieldCodeGeneratorBuilder

        return FieldCodeGeneratorBuilder(self._type_factory, self._context, self._data)

    def _check_optional_field(self, optional):
        if self._context.reached_optional_field and not optional:
            raise RuntimeError("Optional fields may not be followed by non-optional fields.")

    def _generate_switch(self, protocol_switch):
        switch_code_generator = SwitchCodeGenerator(
            get_required_string_attribute(protocol_switch, "field"),
            self._type_factory,
            self._context,
            self._data,
        )

        protocol_cases = protocol_switch.findall("case")

        switch_code_generator.generate_case_data_interface(protocol_cases)
        switch_code_generator.generate_case_data_field()

        reached_optional_field = self._context.reached_optional_field
        reached_dummy = self._context.reached_dummy
        start = True

        for protocol_case in protocol_cases:
            case_context = switch_code_generator.generate_case(protocol_case, start)

            reached_optional_field = reached_optional_field or case_context.reached_optional_field
            reached_dummy = reached_dummy or case_context.reached_dummy
            start = False

        self._context.reached_optional_field = reached_optional_field
        self._context.reached_dummy = reached_dummy

    def _generate_chunked(self, protocol_chunked):
        was_already_enabled = self._context.chunked_reading_enabled
        if not was_already_enabled:
            self._context.chunked_reading_enabled = True
            self._data.deserialize.add_line("reader.chunked_reading_mode = True")
            self._data.serialize.add_line("writer.string_sanitization_mode = True")

        for instruction in protocol_chunked:
            self.generate_instruction(instruction)

        if not was_already_enabled:
            self._context.chunked_reading_enabled = False
            self._data.deserialize.add_line("reader.chunked_reading_mode = False")
            self._data.serialize.add_line("writer.string_sanitization_mode = False")

    def _generate_break(self):
        if not self._context.chunked_reading_enabled:
            raise RuntimeError(
                "Cannot generate a break instruction unless chunked reading is enabled."
            )

        self._context.reached_optional_field = False
        self._context.reached_dummy = False

        self._data.serialize.add_line("writer.add_byte(0xFF)")
        self._data.deserialize.add_line("reader.next_chunk()")



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from typing import Union
from .party_reply_code import PartyReplyCode
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class PartyReplyServerPacket(Packet):
    """
    Failed party invite / join request
    """
    _byte_size: int = 0
    _reply_code: PartyReplyCode
    _reply_code_data: 'PartyReplyServerPacket.ReplyCodeData'

    def __init__(self, *, reply_code: PartyReplyCode, reply_code_data: 'PartyReplyServerPacket.ReplyCodeData' = None):
        """
        Create a new instance of PartyReplyServerPacket.

        Args:
            reply_code (PartyReplyCode): 
            reply_code_data (PartyReplyServerPacket.ReplyCodeData): Data associated with the `reply_code` field.
        """
        self._reply_code = reply_code
        self._reply_code_data = reply_code_data

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def reply_code(self) -> PartyReplyCode:
        return self._reply_code

    @property
    def reply_code_data(self) -> 'PartyReplyServerPacket.ReplyCodeData':
        """
        PartyReplyServerPacket.ReplyCodeData: Data associated with the `reply_code` field.
        """
        return self._reply_code_data

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Party

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Reply

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        PartyReplyServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "PartyReplyServerPacket") -> None:
        """
        Serializes an instance of `PartyReplyServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (PartyReplyServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._reply_code is None:
                raise SerializationError("reply_code must be provided.")
            writer.add_char(int(data._reply_code))
            if data._reply_code == PartyReplyCode.AlreadyInAnotherParty:
                if not isinstance(data._reply_code_data, PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty):
                    raise SerializationError("Expected reply_code_data to be type PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty for reply_code " + PartyReplyCode(data._reply_code).name + ".")
                PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty.serialize(writer, data._reply_code_data)
            elif data._reply_code == PartyReplyCode.AlreadyInYourParty:
                if not isinstance(data._reply_code_data, PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty):
                    raise SerializationError("Expected reply_code_data to be type PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty for reply_code " + PartyReplyCode(data._reply_code).name + ".")
                PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty.serialize(writer, data._reply_code_data)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "PartyReplyServerPacket":
        """
        Deserializes an instance of `PartyReplyServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            PartyReplyServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reply_code = PartyReplyCode(reader.get_char())
            reply_code_data: PartyReplyServerPacket.ReplyCodeData = None
            if reply_code == PartyReplyCode.AlreadyInAnotherParty:
                reply_code_data = PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty.deserialize(reader)
            elif reply_code == PartyReplyCode.AlreadyInYourParty:
                reply_code_data = PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty.deserialize(reader)
            result = PartyReplyServerPacket(reply_code=reply_code, reply_code_data=reply_code_data)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"PartyReplyServerPacket(byte_size={repr(self._byte_size)}, reply_code={repr(self._reply_code)}, reply_code_data={repr(self._reply_code_data)})"

    ReplyCodeData = Union['PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty', 'PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty', None]
    """
    Data associated with different values of the `reply_code` field.
    """

    class ReplyCodeDataAlreadyInAnotherParty:
        """
        Data associated with reply_code value PartyReplyCode.AlreadyInAnotherParty
        """
        _byte_size: int = 0
        _player_name: str

        def __init__(self, *, player_name: str):
            """
            Create a new instance of PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty.

            Args:
                player_name (str): 
            """
            self._player_name = player_name

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def player_name(self) -> str:
            return self._player_name

        @staticmethod
        def serialize(writer: EoWriter, data: "PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty") -> None:
            """
            Serializes an instance of `PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._player_name is None:
                    raise SerializationError("player_name must be provided.")
                writer.add_string(data._player_name)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty":
            """
            Deserializes an instance of `PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                player_name = reader.get_string()
                result = PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty(player_name=player_name)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"PartyReplyServerPacket.ReplyCodeDataAlreadyInAnotherParty(byte_size={repr(self._byte_size)}, player_name={repr(self._player_name)})"

    class ReplyCodeDataAlreadyInYourParty:
        """
        Data associated with reply_code value PartyReplyCode.AlreadyInYourParty
        """
        _byte_size: int = 0
        _player_name: str

        def __init__(self, *, player_name: str):
            """
            Create a new instance of PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty.

            Args:
                player_name (str): 
            """
            self._player_name = player_name

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def player_name(self) -> str:
            return self._player_name

        @staticmethod
        def serialize(writer: EoWriter, data: "PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty") -> None:
            """
            Serializes an instance of `PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._player_name is None:
                    raise SerializationError("player_name must be provided.")
                writer.add_string(data._player_name)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty":
            """
            Deserializes an instance of `PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                player_name = reader.get_string()
                result = PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty(player_name=player_name)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"PartyReplyServerPacket.ReplyCodeDataAlreadyInYourParty(byte_size={repr(self._byte_size)}, player_name={repr(self._player_name)})"



import pytest
from collections import namedtuple
from eolib.data.eo_numeric_limits import THREE_MAX
from eolib.encrypt.server_verification_utils import server_verification_hash

TCase = namedtuple("TCase", ["challenge", "expected_hash"])


def idfn(test_case: TCase) -> str:
    return str(test_case.challenge)


@pytest.mark.parametrize(
    "test_case",
    [
        TCase(0, 114000),
        TCase(1, 115191),
        TCase(2, 229432),
        TCase(5, 613210),
        TCase(12345, 266403),
        TCase(100_000, 145554),
        TCase(5_000_000, 339168),
        TCase(11_092_003, 112773),
        TCase(11_092_004, 112655),
        TCase(11_092_005, 112299),
        TCase(11_092_110, 11016),
        TCase(11_092_111, -2787),
        TCase(11_111_111, 103749),
        TCase(12_345_678, -32046),
        TCase(THREE_MAX - 1, 105960),
    ],
    ids=idfn,
)
def test_server_verification_hash(test_case: TCase):
    challenge_hash = server_verification_hash(test_case.challenge)
    assert challenge_hash == test_case.expected_hash



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from .group_heal_target_player import GroupHealTargetPlayer
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class SpellTargetGroupServerPacket(Packet):
    """
    Nearby player(s) hit by a group heal spell from a player
    """
    _byte_size: int = 0
    _spell_id: int
    _caster_id: int
    _caster_tp: int
    _spell_heal_hp: int
    _players: tuple[GroupHealTargetPlayer, ...]

    def __init__(self, *, spell_id: int, caster_id: int, caster_tp: int, spell_heal_hp: int, players: Iterable[GroupHealTargetPlayer]):
        """
        Create a new instance of SpellTargetGroupServerPacket.

        Args:
            spell_id (int): (Value range is 0-64008.)
            caster_id (int): (Value range is 0-64008.)
            caster_tp (int): (Value range is 0-64008.)
            spell_heal_hp (int): (Value range is 0-64008.)
            players (Iterable[GroupHealTargetPlayer]): 
        """
        self._spell_id = spell_id
        self._caster_id = caster_id
        self._caster_tp = caster_tp
        self._spell_heal_hp = spell_heal_hp
        self._players = tuple(players)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def spell_id(self) -> int:
        return self._spell_id

    @property
    def caster_id(self) -> int:
        return self._caster_id

    @property
    def caster_tp(self) -> int:
        return self._caster_tp

    @property
    def spell_heal_hp(self) -> int:
        return self._spell_heal_hp

    @property
    def players(self) -> tuple[GroupHealTargetPlayer, ...]:
        return self._players

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Spell

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.TargetGroup

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        SpellTargetGroupServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "SpellTargetGroupServerPacket") -> None:
        """
        Serializes an instance of `SpellTargetGroupServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (SpellTargetGroupServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._spell_id is None:
                raise SerializationError("spell_id must be provided.")
            writer.add_short(data._spell_id)
            if data._caster_id is None:
                raise SerializationError("caster_id must be provided.")
            writer.add_short(data._caster_id)
            if data._caster_tp is None:
                raise SerializationError("caster_tp must be provided.")
            writer.add_short(data._caster_tp)
            if data._spell_heal_hp is None:
                raise SerializationError("spell_heal_hp must be provided.")
            writer.add_short(data._spell_heal_hp)
            if data._players is None:
                raise SerializationError("players must be provided.")
            for i in range(len(data._players)):
                GroupHealTargetPlayer.serialize(writer, data._players[i])
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "SpellTargetGroupServerPacket":
        """
        Deserializes an instance of `SpellTargetGroupServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            SpellTargetGroupServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            spell_id = reader.get_short()
            caster_id = reader.get_short()
            caster_tp = reader.get_short()
            spell_heal_hp = reader.get_short()
            players_length = int(reader.remaining / 5)
            players = []
            for i in range(players_length):
                players.append(GroupHealTargetPlayer.deserialize(reader))
            result = SpellTargetGroupServerPacket(spell_id=spell_id, caster_id=caster_id, caster_tp=caster_tp, spell_heal_hp=spell_heal_hp, players=players)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"SpellTargetGroupServerPacket(byte_size={repr(self._byte_size)}, spell_id={repr(self._spell_id)}, caster_id={repr(self._caster_id)}, caster_tp={repr(self._caster_tp)}, spell_heal_hp={repr(self._spell_heal_hp)}, players={repr(self._players)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from .esf_record import EsfRecord
from ...serialization_error import SerializationError
from ....data.eo_writer import EoWriter
from ....data.eo_reader import EoReader

class Esf:
    """
    Endless Skill File
    """
    _byte_size: int = 0
    _rid: tuple[int, ...]
    _total_skills_count: int
    _version: int
    _skills: tuple[EsfRecord, ...]

    def __init__(self, *, rid: Iterable[int], total_skills_count: int, version: int, skills: Iterable[EsfRecord]):
        """
        Create a new instance of Esf.

        Args:
            rid (Iterable[int]): (Length must be `2`.) (Element value range is 0-64008.)
            total_skills_count (int): (Value range is 0-64008.)
            version (int): (Value range is 0-252.)
            skills (Iterable[EsfRecord]): 
        """
        self._rid = tuple(rid)
        self._total_skills_count = total_skills_count
        self._version = version
        self._skills = tuple(skills)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def rid(self) -> tuple[int, ...]:
        return self._rid

    @property
    def total_skills_count(self) -> int:
        return self._total_skills_count

    @property
    def version(self) -> int:
        return self._version

    @property
    def skills(self) -> tuple[EsfRecord, ...]:
        return self._skills

    @staticmethod
    def serialize(writer: EoWriter, data: "Esf") -> None:
        """
        Serializes an instance of `Esf` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (Esf): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.add_fixed_string("ESF", 3, False)
            if data._rid is None:
                raise SerializationError("rid must be provided.")
            if len(data._rid) != 2:
                raise SerializationError(f"Expected length of rid to be exactly 2, got {len(data._rid)}.")
            for i in range(2):
                writer.add_short(data._rid[i])
            if data._total_skills_count is None:
                raise SerializationError("total_skills_count must be provided.")
            writer.add_short(data._total_skills_count)
            if data._version is None:
                raise SerializationError("version must be provided.")
            writer.add_char(data._version)
            if data._skills is None:
                raise SerializationError("skills must be provided.")
            for i in range(len(data._skills)):
                EsfRecord.serialize(writer, data._skills[i])
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "Esf":
        """
        Deserializes an instance of `Esf` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            Esf: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.get_fixed_string(3, False)
            rid = []
            for i in range(2):
                rid.append(reader.get_short())
            total_skills_count = reader.get_short()
            version = reader.get_char()
            skills = []
            while reader.remaining > 0:
                skills.append(EsfRecord.deserialize(reader))
            result = Esf(rid=rid, total_skills_count=total_skills_count, version=version, skills=skills)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"Esf(byte_size={repr(self._byte_size)}, rid={repr(self._rid)}, total_skills_count={repr(self._total_skills_count)}, version={repr(self._version)}, skills={repr(self._skills)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ...protocol_enum_meta import ProtocolEnumMeta

class ItemSubtype(IntEnum, metaclass=ProtocolEnumMeta):
    None_ = 0
    Ranged = 1
    Arrows = 2
    Wings = 3
    Reserved4 = 4



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class DialogReply(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Whether the player has clicked the OK button or a link in a quest dialog
    """
    Ok = 1
    Link = 2



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class DropRecord:
    """
    Record of an item an NPC can drop when killed
    """
    _byte_size: int = 0
    _item_id: int
    _min_amount: int
    _max_amount: int
    _rate: int

    def __init__(self, *, item_id: int, min_amount: int, max_amount: int, rate: int):
        """
        Create a new instance of DropRecord.

        Args:
            item_id (int): (Value range is 0-64008.)
            min_amount (int): (Value range is 0-16194276.)
            max_amount (int): (Value range is 0-16194276.)
            rate (int): Chance (x in 64,000) of the item being dropped (Value range is 0-64008.)
        """
        self._item_id = item_id
        self._min_amount = min_amount
        self._max_amount = max_amount
        self._rate = rate

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def item_id(self) -> int:
        return self._item_id

    @property
    def min_amount(self) -> int:
        return self._min_amount

    @property
    def max_amount(self) -> int:
        return self._max_amount

    @property
    def rate(self) -> int:
        """
        Chance (x in 64,000) of the item being dropped
        """
        return self._rate

    @staticmethod
    def serialize(writer: EoWriter, data: "DropRecord") -> None:
        """
        Serializes an instance of `DropRecord` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (DropRecord): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._item_id is None:
                raise SerializationError("item_id must be provided.")
            writer.add_short(data._item_id)
            if data._min_amount is None:
                raise SerializationError("min_amount must be provided.")
            writer.add_three(data._min_amount)
            if data._max_amount is None:
                raise SerializationError("max_amount must be provided.")
            writer.add_three(data._max_amount)
            if data._rate is None:
                raise SerializationError("rate must be provided.")
            writer.add_short(data._rate)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "DropRecord":
        """
        Deserializes an instance of `DropRecord` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            DropRecord: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            item_id = reader.get_short()
            min_amount = reader.get_three()
            max_amount = reader.get_three()
            rate = reader.get_short()
            result = DropRecord(item_id=item_id, min_amount=min_amount, max_amount=max_amount, rate=rate)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"DropRecord(byte_size={repr(self._byte_size)}, item_id={repr(self._item_id)}, min_amount={repr(self._min_amount)}, max_amount={repr(self._max_amount)}, rate={repr(self._rate)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from typing import Union
from collections.abc import Iterable
from .character_selection_list_entry import CharacterSelectionListEntry
from .character_reply import CharacterReply
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class CharacterReplyServerPacket(Packet):
    """
    Reply to client Character-family packets
    """
    _byte_size: int = 0
    _reply_code: CharacterReply
    _reply_code_data: 'CharacterReplyServerPacket.ReplyCodeData'

    def __init__(self, *, reply_code: CharacterReply, reply_code_data: 'CharacterReplyServerPacket.ReplyCodeData' = None):
        """
        Create a new instance of CharacterReplyServerPacket.

        Args:
            reply_code (CharacterReply): Sometimes a CharacterReply code, sometimes a session ID for character creation
            reply_code_data (CharacterReplyServerPacket.ReplyCodeData): Data associated with the `reply_code` field.
        """
        self._reply_code = reply_code
        self._reply_code_data = reply_code_data

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def reply_code(self) -> CharacterReply:
        """
        Sometimes a CharacterReply code, sometimes a session ID for character creation
        """
        return self._reply_code

    @property
    def reply_code_data(self) -> 'CharacterReplyServerPacket.ReplyCodeData':
        """
        CharacterReplyServerPacket.ReplyCodeData: Data associated with the `reply_code` field.
        """
        return self._reply_code_data

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Character

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Reply

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        CharacterReplyServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "CharacterReplyServerPacket") -> None:
        """
        Serializes an instance of `CharacterReplyServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (CharacterReplyServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.string_sanitization_mode = True
            if data._reply_code is None:
                raise SerializationError("reply_code must be provided.")
            writer.add_short(int(data._reply_code))
            if data._reply_code == 0:
                if data._reply_code_data is not None:
                    raise SerializationError("Expected reply_code_data to be None for reply_code " + CharacterReply(data._reply_code).name + ".")
            elif data._reply_code == CharacterReply.Exists:
                if not isinstance(data._reply_code_data, CharacterReplyServerPacket.ReplyCodeDataExists):
                    raise SerializationError("Expected reply_code_data to be type CharacterReplyServerPacket.ReplyCodeDataExists for reply_code " + CharacterReply(data._reply_code).name + ".")
                CharacterReplyServerPacket.ReplyCodeDataExists.serialize(writer, data._reply_code_data)
            elif data._reply_code == CharacterReply.Full:
                if not isinstance(data._reply_code_data, CharacterReplyServerPacket.ReplyCodeDataFull):
                    raise SerializationError("Expected reply_code_data to be type CharacterReplyServerPacket.ReplyCodeDataFull for reply_code " + CharacterReply(data._reply_code).name + ".")
                CharacterReplyServerPacket.ReplyCodeDataFull.serialize(writer, data._reply_code_data)
            elif data._reply_code == CharacterReply.Full3:
                if not isinstance(data._reply_code_data, CharacterReplyServerPacket.ReplyCodeDataFull3):
                    raise SerializationError("Expected reply_code_data to be type CharacterReplyServerPacket.ReplyCodeDataFull3 for reply_code " + CharacterReply(data._reply_code).name + ".")
                CharacterReplyServerPacket.ReplyCodeDataFull3.serialize(writer, data._reply_code_data)
            elif data._reply_code == CharacterReply.NotApproved:
                if not isinstance(data._reply_code_data, CharacterReplyServerPacket.ReplyCodeDataNotApproved):
                    raise SerializationError("Expected reply_code_data to be type CharacterReplyServerPacket.ReplyCodeDataNotApproved for reply_code " + CharacterReply(data._reply_code).name + ".")
                CharacterReplyServerPacket.ReplyCodeDataNotApproved.serialize(writer, data._reply_code_data)
            elif data._reply_code == CharacterReply.Ok:
                if not isinstance(data._reply_code_data, CharacterReplyServerPacket.ReplyCodeDataOk):
                    raise SerializationError("Expected reply_code_data to be type CharacterReplyServerPacket.ReplyCodeDataOk for reply_code " + CharacterReply(data._reply_code).name + ".")
                CharacterReplyServerPacket.ReplyCodeDataOk.serialize(writer, data._reply_code_data)
            elif data._reply_code == CharacterReply.Deleted:
                if not isinstance(data._reply_code_data, CharacterReplyServerPacket.ReplyCodeDataDeleted):
                    raise SerializationError("Expected reply_code_data to be type CharacterReplyServerPacket.ReplyCodeDataDeleted for reply_code " + CharacterReply(data._reply_code).name + ".")
                CharacterReplyServerPacket.ReplyCodeDataDeleted.serialize(writer, data._reply_code_data)
            elif data._reply_code == 7:
                if data._reply_code_data is not None:
                    raise SerializationError("Expected reply_code_data to be None for reply_code " + CharacterReply(data._reply_code).name + ".")
            elif data._reply_code == 8:
                if data._reply_code_data is not None:
                    raise SerializationError("Expected reply_code_data to be None for reply_code " + CharacterReply(data._reply_code).name + ".")
            elif data._reply_code == 9:
                if data._reply_code_data is not None:
                    raise SerializationError("Expected reply_code_data to be None for reply_code " + CharacterReply(data._reply_code).name + ".")
            else:
                if not isinstance(data._reply_code_data, CharacterReplyServerPacket.ReplyCodeDataDefault):
                    raise SerializationError("Expected reply_code_data to be type CharacterReplyServerPacket.ReplyCodeDataDefault for reply_code " + CharacterReply(data._reply_code).name + ".")
                CharacterReplyServerPacket.ReplyCodeDataDefault.serialize(writer, data._reply_code_data)
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "CharacterReplyServerPacket":
        """
        Deserializes an instance of `CharacterReplyServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            CharacterReplyServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.chunked_reading_mode = True
            reply_code = CharacterReply(reader.get_short())
            reply_code_data: CharacterReplyServerPacket.ReplyCodeData = None
            if reply_code == 0:
                reply_code_data = None
            elif reply_code == CharacterReply.Exists:
                reply_code_data = CharacterReplyServerPacket.ReplyCodeDataExists.deserialize(reader)
            elif reply_code == CharacterReply.Full:
                reply_code_data = CharacterReplyServerPacket.ReplyCodeDataFull.deserialize(reader)
            elif reply_code == CharacterReply.Full3:
                reply_code_data = CharacterReplyServerPacket.ReplyCodeDataFull3.deserialize(reader)
            elif reply_code == CharacterReply.NotApproved:
                reply_code_data = CharacterReplyServerPacket.ReplyCodeDataNotApproved.deserialize(reader)
            elif reply_code == CharacterReply.Ok:
                reply_code_data = CharacterReplyServerPacket.ReplyCodeDataOk.deserialize(reader)
            elif reply_code == CharacterReply.Deleted:
                reply_code_data = CharacterReplyServerPacket.ReplyCodeDataDeleted.deserialize(reader)
            elif reply_code == 7:
                reply_code_data = None
            elif reply_code == 8:
                reply_code_data = None
            elif reply_code == 9:
                reply_code_data = None
            else:
                reply_code_data = CharacterReplyServerPacket.ReplyCodeDataDefault.deserialize(reader)
            reader.chunked_reading_mode = False
            result = CharacterReplyServerPacket(reply_code=reply_code, reply_code_data=reply_code_data)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"CharacterReplyServerPacket(byte_size={repr(self._byte_size)}, reply_code={repr(self._reply_code)}, reply_code_data={repr(self._reply_code_data)})"

    ReplyCodeData = Union['CharacterReplyServerPacket.ReplyCodeDataExists', 'CharacterReplyServerPacket.ReplyCodeDataFull', 'CharacterReplyServerPacket.ReplyCodeDataFull3', 'CharacterReplyServerPacket.ReplyCodeDataNotApproved', 'CharacterReplyServerPacket.ReplyCodeDataOk', 'CharacterReplyServerPacket.ReplyCodeDataDeleted', 'CharacterReplyServerPacket.ReplyCodeDataDefault', None]
    """
    Data associated with different values of the `reply_code` field.
    """

    class ReplyCodeDataExists:
        """
        Data associated with reply_code value CharacterReply.Exists
        """
        _byte_size: int = 0

        def __init__(self):
            """
            Create a new instance of CharacterReplyServerPacket.ReplyCodeDataExists.
            """

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size


        @staticmethod
        def serialize(writer: EoWriter, data: "CharacterReplyServerPacket.ReplyCodeDataExists") -> None:
            """
            Serializes an instance of `CharacterReplyServerPacket.ReplyCodeDataExists` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (CharacterReplyServerPacket.ReplyCodeDataExists): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                writer.add_string("NO")
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "CharacterReplyServerPacket.ReplyCodeDataExists":
            """
            Deserializes an instance of `CharacterReplyServerPacket.ReplyCodeDataExists` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                CharacterReplyServerPacket.ReplyCodeDataExists: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                reader.get_string()
                result = CharacterReplyServerPacket.ReplyCodeDataExists()
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"CharacterReplyServerPacket.ReplyCodeDataExists(byte_size={repr(self._byte_size)})"

    class ReplyCodeDataFull:
        """
        Data associated with reply_code value CharacterReply.Full
        """
        _byte_size: int = 0

        def __init__(self):
            """
            Create a new instance of CharacterReplyServerPacket.ReplyCodeDataFull.
            """

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size


        @staticmethod
        def serialize(writer: EoWriter, data: "CharacterReplyServerPacket.ReplyCodeDataFull") -> None:
            """
            Serializes an instance of `CharacterReplyServerPacket.ReplyCodeDataFull` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (CharacterReplyServerPacket.ReplyCodeDataFull): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                writer.add_string("NO")
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "CharacterReplyServerPacket.ReplyCodeDataFull":
            """
            Deserializes an instance of `CharacterReplyServerPacket.ReplyCodeDataFull` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                CharacterReplyServerPacket.ReplyCodeDataFull: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                reader.get_string()
                result = CharacterReplyServerPacket.ReplyCodeDataFull()
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"CharacterReplyServerPacket.ReplyCodeDataFull(byte_size={repr(self._byte_size)})"

    class ReplyCodeDataFull3:
        """
        Data associated with reply_code value CharacterReply.Full3
        """
        _byte_size: int = 0

        def __init__(self):
            """
            Create a new instance of CharacterReplyServerPacket.ReplyCodeDataFull3.
            """

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size


        @staticmethod
        def serialize(writer: EoWriter, data: "CharacterReplyServerPacket.ReplyCodeDataFull3") -> None:
            """
            Serializes an instance of `CharacterReplyServerPacket.ReplyCodeDataFull3` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (CharacterReplyServerPacket.ReplyCodeDataFull3): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                writer.add_string("NO")
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "CharacterReplyServerPacket.ReplyCodeDataFull3":
            """
            Deserializes an instance of `CharacterReplyServerPacket.ReplyCodeDataFull3` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                CharacterReplyServerPacket.ReplyCodeDataFull3: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                reader.get_string()
                result = CharacterReplyServerPacket.ReplyCodeDataFull3()
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"CharacterReplyServerPacket.ReplyCodeDataFull3(byte_size={repr(self._byte_size)})"

    class ReplyCodeDataNotApproved:
        """
        Data associated with reply_code value CharacterReply.NotApproved
        """
        _byte_size: int = 0

        def __init__(self):
            """
            Create a new instance of CharacterReplyServerPacket.ReplyCodeDataNotApproved.
            """

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size


        @staticmethod
        def serialize(writer: EoWriter, data: "CharacterReplyServerPacket.ReplyCodeDataNotApproved") -> None:
            """
            Serializes an instance of `CharacterReplyServerPacket.ReplyCodeDataNotApproved` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (CharacterReplyServerPacket.ReplyCodeDataNotApproved): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                writer.add_string("NO")
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "CharacterReplyServerPacket.ReplyCodeDataNotApproved":
            """
            Deserializes an instance of `CharacterReplyServerPacket.ReplyCodeDataNotApproved` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                CharacterReplyServerPacket.ReplyCodeDataNotApproved: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                reader.get_string()
                result = CharacterReplyServerPacket.ReplyCodeDataNotApproved()
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"CharacterReplyServerPacket.ReplyCodeDataNotApproved(byte_size={repr(self._byte_size)})"

    class ReplyCodeDataOk:
        """
        Data associated with reply_code value CharacterReply.Ok
        """
        _byte_size: int = 0
        _characters_count: int
        _characters: tuple[CharacterSelectionListEntry, ...]

        def __init__(self, *, characters: Iterable[CharacterSelectionListEntry]):
            """
            Create a new instance of CharacterReplyServerPacket.ReplyCodeDataOk.

            Args:
                characters (Iterable[CharacterSelectionListEntry]): (Length must be 252 or less.)
            """
            self._characters = tuple(characters)
            self._characters_count = len(self._characters)

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def characters(self) -> tuple[CharacterSelectionListEntry, ...]:
            return self._characters

        @staticmethod
        def serialize(writer: EoWriter, data: "CharacterReplyServerPacket.ReplyCodeDataOk") -> None:
            """
            Serializes an instance of `CharacterReplyServerPacket.ReplyCodeDataOk` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (CharacterReplyServerPacket.ReplyCodeDataOk): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._characters_count is None:
                    raise SerializationError("characters_count must be provided.")
                writer.add_char(data._characters_count)
                writer.add_char(0)
                writer.add_byte(0xFF)
                if data._characters is None:
                    raise SerializationError("characters must be provided.")
                if len(data._characters) > 252:
                    raise SerializationError(f"Expected length of characters to be 252 or less, got {len(data._characters)}.")
                for i in range(data._characters_count):
                    CharacterSelectionListEntry.serialize(writer, data._characters[i])
                    writer.add_byte(0xFF)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "CharacterReplyServerPacket.ReplyCodeDataOk":
            """
            Deserializes an instance of `CharacterReplyServerPacket.ReplyCodeDataOk` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                CharacterReplyServerPacket.ReplyCodeDataOk: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                characters_count = reader.get_char()
                reader.get_char()
                reader.next_chunk()
                characters = []
                for i in range(characters_count):
                    characters.append(CharacterSelectionListEntry.deserialize(reader))
                    reader.next_chunk()
                result = CharacterReplyServerPacket.ReplyCodeDataOk(characters=characters)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"CharacterReplyServerPacket.ReplyCodeDataOk(byte_size={repr(self._byte_size)}, characters={repr(self._characters)})"

    class ReplyCodeDataDeleted:
        """
        Data associated with reply_code value CharacterReply.Deleted
        """
        _byte_size: int = 0
        _characters_count: int
        _characters: tuple[CharacterSelectionListEntry, ...]

        def __init__(self, *, characters: Iterable[CharacterSelectionListEntry]):
            """
            Create a new instance of CharacterReplyServerPacket.ReplyCodeDataDeleted.

            Args:
                characters (Iterable[CharacterSelectionListEntry]): (Length must be 252 or less.)
            """
            self._characters = tuple(characters)
            self._characters_count = len(self._characters)

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def characters(self) -> tuple[CharacterSelectionListEntry, ...]:
            return self._characters

        @staticmethod
        def serialize(writer: EoWriter, data: "CharacterReplyServerPacket.ReplyCodeDataDeleted") -> None:
            """
            Serializes an instance of `CharacterReplyServerPacket.ReplyCodeDataDeleted` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (CharacterReplyServerPacket.ReplyCodeDataDeleted): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._characters_count is None:
                    raise SerializationError("characters_count must be provided.")
                writer.add_char(data._characters_count)
                writer.add_byte(0xFF)
                if data._characters is None:
                    raise SerializationError("characters must be provided.")
                if len(data._characters) > 252:
                    raise SerializationError(f"Expected length of characters to be 252 or less, got {len(data._characters)}.")
                for i in range(data._characters_count):
                    CharacterSelectionListEntry.serialize(writer, data._characters[i])
                    writer.add_byte(0xFF)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "CharacterReplyServerPacket.ReplyCodeDataDeleted":
            """
            Deserializes an instance of `CharacterReplyServerPacket.ReplyCodeDataDeleted` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                CharacterReplyServerPacket.ReplyCodeDataDeleted: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                characters_count = reader.get_char()
                reader.next_chunk()
                characters = []
                for i in range(characters_count):
                    characters.append(CharacterSelectionListEntry.deserialize(reader))
                    reader.next_chunk()
                result = CharacterReplyServerPacket.ReplyCodeDataDeleted(characters=characters)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"CharacterReplyServerPacket.ReplyCodeDataDeleted(byte_size={repr(self._byte_size)}, characters={repr(self._characters)})"

    class ReplyCodeDataDefault:
        """
        Default data associated with reply_code

        In this case (reply_code &gt; 9), reply_code is a session ID for character creation
        """
        _byte_size: int = 0

        def __init__(self):
            """
            Create a new instance of CharacterReplyServerPacket.ReplyCodeDataDefault.
            """

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size


        @staticmethod
        def serialize(writer: EoWriter, data: "CharacterReplyServerPacket.ReplyCodeDataDefault") -> None:
            """
            Serializes an instance of `CharacterReplyServerPacket.ReplyCodeDataDefault` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (CharacterReplyServerPacket.ReplyCodeDataDefault): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                writer.add_string("OK")
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "CharacterReplyServerPacket.ReplyCodeDataDefault":
            """
            Deserializes an instance of `CharacterReplyServerPacket.ReplyCodeDataDefault` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                CharacterReplyServerPacket.ReplyCodeDataDefault: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                reader.get_string()
                result = CharacterReplyServerPacket.ReplyCodeDataDefault()
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"CharacterReplyServerPacket.ReplyCodeDataDefault(byte_size={repr(self._byte_size)})"



import pytest
from eolib.data.eo_numeric_limits import CHAR_MAX, SHORT_MAX, THREE_MAX, INT_MAX
from eolib.data.eo_writer import EoWriter


def test_add_byte():
    writer = EoWriter()
    writer.add_byte(0x00)
    assert writer.to_bytearray() == bytearray([0x00])


def test_add_bytes():
    writer = EoWriter()
    writer.add_bytes(bytes([0x00, 0xFF]))
    assert writer.to_bytearray() == bytearray([0x00, 0xFF])


def test_add_char():
    writer = EoWriter()
    writer.add_char(123)
    assert writer.to_bytearray() == bytearray([0x7C])


def test_add_short():
    writer = EoWriter()
    writer.add_short(12345)
    assert writer.to_bytearray() == bytearray([0xCA, 0x31])


def test_add_three():
    writer = EoWriter()
    writer.add_three(10000000)
    assert writer.to_bytearray() == bytearray([0xB0, 0x3A, 0x9D])


def test_add_int():
    writer = EoWriter()
    writer.add_int(2048576040)
    assert writer.to_bytearray() == bytearray([0x7F, 0x7F, 0x7F, 0x7F])


def test_add_string():
    writer = EoWriter()
    writer.add_string("foo")
    assert writer.to_bytearray() == b'foo'


def test_add_fixed_string():
    writer = EoWriter()
    writer.add_fixed_string("bar", 3)
    assert writer.to_bytearray() == b'bar'


def test_add_padded_fixed_string():
    writer = EoWriter()
    writer.add_fixed_string("bar", 6, True)
    assert writer.to_bytearray() == b'bar\xff\xff\xff'


def test_add_padded_with_perfect_fit_fixed_string():
    writer = EoWriter()
    writer.add_fixed_string("bar", 3, True)
    assert writer.to_bytearray() == b'bar'


def test_add_encoded_string():
    writer = EoWriter()
    writer.add_encoded_string("foo")
    assert writer.to_bytearray() == b'^0g'


def test_add_fixed_encoded_string():
    writer = EoWriter()
    writer.add_fixed_encoded_string("bar", 3)
    assert writer.to_bytearray() == b'[>k'


def test_add_padded_fixed_encoded_string():
    writer = EoWriter()
    writer.add_fixed_encoded_string("bar", 6, True)
    assert writer.to_bytearray() == b'\xff\xff\xff-l='


def test_add_padded_with_perfect_fit_fixed_encoded_string():
    writer = EoWriter()
    writer.add_fixed_encoded_string("bar", 3, True)
    assert writer.to_bytearray() == b'[>k'


def test_add_sanitized_string():
    writer = EoWriter()
    writer.string_sanitization_mode = True
    writer.add_string("aÿz")
    assert writer.to_bytearray() == b'ayz'


def test_add_sanitized_fixed_string():
    writer = EoWriter()
    writer.string_sanitization_mode = True
    writer.add_fixed_string("aÿz", 3)
    assert writer.to_bytearray() == b'ayz'


def test_add_sanitized_padded_fixed_string():
    writer = EoWriter()
    writer.string_sanitization_mode = True
    writer.add_fixed_string("aÿz", 6, True)
    assert writer.to_bytearray() == b'ayz\xff\xff\xff'


def test_add_sanitized_encoded_string():
    writer = EoWriter()
    writer.string_sanitization_mode = True
    writer.add_encoded_string("aÿz")
    assert writer.to_bytearray() == b'S&l'


def test_add_sanitized_fixed_encoded_string():
    writer = EoWriter()
    writer.string_sanitization_mode = True
    writer.add_fixed_encoded_string("aÿz", 3)
    assert writer.to_bytearray() == b'S&l'


def test_add_sanitized_padded_fixed_encoded_string():
    writer = EoWriter()
    writer.string_sanitization_mode = True
    writer.add_fixed_encoded_string("aÿz", 6, True)
    assert writer.to_bytearray() == b'\xff\xff\xff%T>'


def test_add_numbers_on_boundary():
    writer = EoWriter()
    writer.add_byte(0xFF)
    writer.add_char(CHAR_MAX - 1)
    writer.add_short(SHORT_MAX - 1)
    writer.add_three(THREE_MAX - 1)
    writer.add_int(INT_MAX - 1)


def test_add_numbers_exceeding_limit():
    writer = EoWriter()
    with pytest.raises(ValueError):
        writer.add_byte(256)
    with pytest.raises(ValueError):
        writer.add_char(256)
    with pytest.raises(ValueError):
        writer.add_short(65536)
    with pytest.raises(ValueError):
        writer.add_three(16777216)
    with pytest.raises(ValueError):
        writer.add_int(4294967296)


def test_add_fixed_string_with_incorrect_length():
    writer = EoWriter()
    with pytest.raises(ValueError):
        writer.add_fixed_string("foo", 2)
    with pytest.raises(ValueError):
        writer.add_fixed_string("foo", 2, True)
    with pytest.raises(ValueError):
        writer.add_fixed_string("foo", 4)
    with pytest.raises(ValueError):
        writer.add_fixed_encoded_string("foo", 2)
    with pytest.raises(ValueError):
        writer.add_fixed_encoded_string("foo", 2, True)
    with pytest.raises(ValueError):
        writer.add_fixed_encoded_string("foo", 4)


def test_get_string_sanitization_mode():
    writer = EoWriter()
    assert not writer.string_sanitization_mode
    writer.string_sanitization_mode = True
    assert writer.string_sanitization_mode


def test_len():
    writer = EoWriter()
    assert len(writer) == 0

    writer.add_string("Lorem ipsum dolor sit amet")
    assert len(writer) == 26

    for _ in range(27, 101):
        writer.add_byte(0xFF)
    assert len(writer) == 100



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from typing import cast
from typing import Optional
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class ChestCloseServerPacket(Packet):
    """
    Reply to trying to interact with a locked or "broken" chest.
    The official client assumes a broken chest if the packet is under 2 bytes in length.
    """
    _byte_size: int = 0
    _key: Optional[int]

    def __init__(self, *, key: Optional[int] = None):
        """
        Create a new instance of ChestCloseServerPacket.

        Args:
            key (Optional[int]): Sent if the player is trying to interact with a locked chest (Value range is 0-64008.)
        """
        self._key = key

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def key(self) -> Optional[int]:
        """
        Sent if the player is trying to interact with a locked chest
        """
        return self._key

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Chest

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Close

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        ChestCloseServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "ChestCloseServerPacket") -> None:
        """
        Serializes an instance of `ChestCloseServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (ChestCloseServerPacket): The data to serialize.
        """
        old_writer_length: int = len(writer)
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            reached_missing_optional = data._key is None
            if not reached_missing_optional:
                writer.add_short(cast(int, data._key))
            if len(writer) == old_writer_length:
                writer.add_string("N")
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "ChestCloseServerPacket":
        """
        Deserializes an instance of `ChestCloseServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            ChestCloseServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            key: Optional[int] = None
            if reader.remaining > 0:
                key = reader.get_short()
            if reader.position == reader_start_position:
                reader.get_string()
            result = ChestCloseServerPacket(key=key)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"ChestCloseServerPacket(byte_size={repr(self._byte_size)}, key={repr(self._key)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class JukeboxUseClientPacket(Packet):
    """
    Playing a note with the bard skill
    """
    _byte_size: int = 0
    _instrument_id: int
    _note_id: int

    def __init__(self, *, instrument_id: int, note_id: int):
        """
        Create a new instance of JukeboxUseClientPacket.

        Args:
            instrument_id (int): (Value range is 0-252.)
            note_id (int): (Value range is 0-252.)
        """
        self._instrument_id = instrument_id
        self._note_id = note_id

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def instrument_id(self) -> int:
        return self._instrument_id

    @property
    def note_id(self) -> int:
        return self._note_id

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Jukebox

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Use

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        JukeboxUseClientPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "JukeboxUseClientPacket") -> None:
        """
        Serializes an instance of `JukeboxUseClientPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (JukeboxUseClientPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._instrument_id is None:
                raise SerializationError("instrument_id must be provided.")
            writer.add_char(data._instrument_id)
            if data._note_id is None:
                raise SerializationError("note_id must be provided.")
            writer.add_char(data._note_id)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "JukeboxUseClientPacket":
        """
        Deserializes an instance of `JukeboxUseClientPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            JukeboxUseClientPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            instrument_id = reader.get_char()
            note_id = reader.get_char()
            result = JukeboxUseClientPacket(instrument_id=instrument_id, note_id=note_id)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"JukeboxUseClientPacket(byte_size={repr(self._byte_size)}, instrument_id={repr(self._instrument_id)}, note_id={repr(self._note_id)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ...direction import Direction
from ...coords import Coords
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class ChairPlayerServerPacket(Packet):
    """
    Nearby player sitting on a chair
    """
    _byte_size: int = 0
    _player_id: int
    _coords: Coords
    _direction: Direction

    def __init__(self, *, player_id: int, coords: Coords, direction: Direction):
        """
        Create a new instance of ChairPlayerServerPacket.

        Args:
            player_id (int): (Value range is 0-64008.)
            coords (Coords): 
            direction (Direction): 
        """
        self._player_id = player_id
        self._coords = coords
        self._direction = direction

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def player_id(self) -> int:
        return self._player_id

    @property
    def coords(self) -> Coords:
        return self._coords

    @property
    def direction(self) -> Direction:
        return self._direction

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Chair

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Player

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        ChairPlayerServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "ChairPlayerServerPacket") -> None:
        """
        Serializes an instance of `ChairPlayerServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (ChairPlayerServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._player_id is None:
                raise SerializationError("player_id must be provided.")
            writer.add_short(data._player_id)
            if data._coords is None:
                raise SerializationError("coords must be provided.")
            Coords.serialize(writer, data._coords)
            if data._direction is None:
                raise SerializationError("direction must be provided.")
            writer.add_char(int(data._direction))
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "ChairPlayerServerPacket":
        """
        Deserializes an instance of `ChairPlayerServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            ChairPlayerServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            player_id = reader.get_short()
            coords = Coords.deserialize(reader)
            direction = Direction(reader.get_char())
            result = ChairPlayerServerPacket(player_id=player_id, coords=coords, direction=direction)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"ChairPlayerServerPacket(byte_size={repr(self._byte_size)}, player_id={repr(self._player_id)}, coords={repr(self._coords)}, direction={repr(self._direction)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from typing import Union
from .map_effect import MapEffect
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class EffectUseServerPacket(Packet):
    """
    Map effect
    """
    _byte_size: int = 0
    _effect: MapEffect
    _effect_data: 'EffectUseServerPacket.EffectData'

    def __init__(self, *, effect: MapEffect, effect_data: 'EffectUseServerPacket.EffectData' = None):
        """
        Create a new instance of EffectUseServerPacket.

        Args:
            effect (MapEffect): 
            effect_data (EffectUseServerPacket.EffectData): Data associated with the `effect` field.
        """
        self._effect = effect
        self._effect_data = effect_data

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def effect(self) -> MapEffect:
        return self._effect

    @property
    def effect_data(self) -> 'EffectUseServerPacket.EffectData':
        """
        EffectUseServerPacket.EffectData: Data associated with the `effect` field.
        """
        return self._effect_data

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Effect

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Use

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        EffectUseServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "EffectUseServerPacket") -> None:
        """
        Serializes an instance of `EffectUseServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (EffectUseServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._effect is None:
                raise SerializationError("effect must be provided.")
            writer.add_char(int(data._effect))
            if data._effect == MapEffect.Quake:
                if not isinstance(data._effect_data, EffectUseServerPacket.EffectDataQuake):
                    raise SerializationError("Expected effect_data to be type EffectUseServerPacket.EffectDataQuake for effect " + MapEffect(data._effect).name + ".")
                EffectUseServerPacket.EffectDataQuake.serialize(writer, data._effect_data)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "EffectUseServerPacket":
        """
        Deserializes an instance of `EffectUseServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            EffectUseServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            effect = MapEffect(reader.get_char())
            effect_data: EffectUseServerPacket.EffectData = None
            if effect == MapEffect.Quake:
                effect_data = EffectUseServerPacket.EffectDataQuake.deserialize(reader)
            result = EffectUseServerPacket(effect=effect, effect_data=effect_data)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"EffectUseServerPacket(byte_size={repr(self._byte_size)}, effect={repr(self._effect)}, effect_data={repr(self._effect_data)})"

    EffectData = Union['EffectUseServerPacket.EffectDataQuake', None]
    """
    Data associated with different values of the `effect` field.
    """

    class EffectDataQuake:
        """
        Data associated with effect value MapEffect.Quake
        """
        _byte_size: int = 0
        _quake_strength: int

        def __init__(self, *, quake_strength: int):
            """
            Create a new instance of EffectUseServerPacket.EffectDataQuake.

            Args:
                quake_strength (int): (Value range is 0-252.)
            """
            self._quake_strength = quake_strength

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def quake_strength(self) -> int:
            return self._quake_strength

        @staticmethod
        def serialize(writer: EoWriter, data: "EffectUseServerPacket.EffectDataQuake") -> None:
            """
            Serializes an instance of `EffectUseServerPacket.EffectDataQuake` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (EffectUseServerPacket.EffectDataQuake): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._quake_strength is None:
                    raise SerializationError("quake_strength must be provided.")
                writer.add_char(data._quake_strength)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "EffectUseServerPacket.EffectDataQuake":
            """
            Deserializes an instance of `EffectUseServerPacket.EffectDataQuake` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                EffectUseServerPacket.EffectDataQuake: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                quake_strength = reader.get_char()
                result = EffectUseServerPacket.EffectDataQuake(quake_strength=quake_strength)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"EffectUseServerPacket.EffectDataQuake(byte_size={repr(self._byte_size)}, quake_strength={repr(self._quake_strength)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class AdminMessageType(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Type of message sent to admins via the Help menu
    """
    Message = 1
    Report = 2



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from typing import Union
from .dialog_reply import DialogReply
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class QuestAcceptClientPacket(Packet):
    """
    Response to a quest NPC dialog
    """
    _byte_size: int = 0
    _session_id: int
    _dialog_id: int
    _quest_id: int
    _npc_index: int
    _reply_type: DialogReply
    _reply_type_data: 'QuestAcceptClientPacket.ReplyTypeData'

    def __init__(self, *, session_id: int, dialog_id: int, quest_id: int, npc_index: int, reply_type: DialogReply, reply_type_data: 'QuestAcceptClientPacket.ReplyTypeData' = None):
        """
        Create a new instance of QuestAcceptClientPacket.

        Args:
            session_id (int): (Value range is 0-64008.)
            dialog_id (int): (Value range is 0-64008.)
            quest_id (int): (Value range is 0-64008.)
            npc_index (int): (Value range is 0-64008.)
            reply_type (DialogReply): 
            reply_type_data (QuestAcceptClientPacket.ReplyTypeData): Data associated with the `reply_type` field.
        """
        self._session_id = session_id
        self._dialog_id = dialog_id
        self._quest_id = quest_id
        self._npc_index = npc_index
        self._reply_type = reply_type
        self._reply_type_data = reply_type_data

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def session_id(self) -> int:
        return self._session_id

    @property
    def dialog_id(self) -> int:
        return self._dialog_id

    @property
    def quest_id(self) -> int:
        return self._quest_id

    @property
    def npc_index(self) -> int:
        return self._npc_index

    @property
    def reply_type(self) -> DialogReply:
        return self._reply_type

    @property
    def reply_type_data(self) -> 'QuestAcceptClientPacket.ReplyTypeData':
        """
        QuestAcceptClientPacket.ReplyTypeData: Data associated with the `reply_type` field.
        """
        return self._reply_type_data

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Quest

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Accept

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        QuestAcceptClientPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "QuestAcceptClientPacket") -> None:
        """
        Serializes an instance of `QuestAcceptClientPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (QuestAcceptClientPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._session_id is None:
                raise SerializationError("session_id must be provided.")
            writer.add_short(data._session_id)
            if data._dialog_id is None:
                raise SerializationError("dialog_id must be provided.")
            writer.add_short(data._dialog_id)
            if data._quest_id is None:
                raise SerializationError("quest_id must be provided.")
            writer.add_short(data._quest_id)
            if data._npc_index is None:
                raise SerializationError("npc_index must be provided.")
            writer.add_short(data._npc_index)
            if data._reply_type is None:
                raise SerializationError("reply_type must be provided.")
            writer.add_char(int(data._reply_type))
            if data._reply_type == DialogReply.Ok:
                if not isinstance(data._reply_type_data, QuestAcceptClientPacket.ReplyTypeDataOk):
                    raise SerializationError("Expected reply_type_data to be type QuestAcceptClientPacket.ReplyTypeDataOk for reply_type " + DialogReply(data._reply_type).name + ".")
                QuestAcceptClientPacket.ReplyTypeDataOk.serialize(writer, data._reply_type_data)
            elif data._reply_type == DialogReply.Link:
                if not isinstance(data._reply_type_data, QuestAcceptClientPacket.ReplyTypeDataLink):
                    raise SerializationError("Expected reply_type_data to be type QuestAcceptClientPacket.ReplyTypeDataLink for reply_type " + DialogReply(data._reply_type).name + ".")
                QuestAcceptClientPacket.ReplyTypeDataLink.serialize(writer, data._reply_type_data)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "QuestAcceptClientPacket":
        """
        Deserializes an instance of `QuestAcceptClientPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            QuestAcceptClientPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            session_id = reader.get_short()
            dialog_id = reader.get_short()
            quest_id = reader.get_short()
            npc_index = reader.get_short()
            reply_type = DialogReply(reader.get_char())
            reply_type_data: QuestAcceptClientPacket.ReplyTypeData = None
            if reply_type == DialogReply.Ok:
                reply_type_data = QuestAcceptClientPacket.ReplyTypeDataOk.deserialize(reader)
            elif reply_type == DialogReply.Link:
                reply_type_data = QuestAcceptClientPacket.ReplyTypeDataLink.deserialize(reader)
            result = QuestAcceptClientPacket(session_id=session_id, dialog_id=dialog_id, quest_id=quest_id, npc_index=npc_index, reply_type=reply_type, reply_type_data=reply_type_data)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"QuestAcceptClientPacket(byte_size={repr(self._byte_size)}, session_id={repr(self._session_id)}, dialog_id={repr(self._dialog_id)}, quest_id={repr(self._quest_id)}, npc_index={repr(self._npc_index)}, reply_type={repr(self._reply_type)}, reply_type_data={repr(self._reply_type_data)})"

    ReplyTypeData = Union['QuestAcceptClientPacket.ReplyTypeDataOk', 'QuestAcceptClientPacket.ReplyTypeDataLink', None]
    """
    Data associated with different values of the `reply_type` field.
    """

    class ReplyTypeDataOk:
        """
        Data associated with reply_type value DialogReply.Ok
        """
        _byte_size: int = 0

        def __init__(self):
            """
            Create a new instance of QuestAcceptClientPacket.ReplyTypeDataOk.
            """

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size


        @staticmethod
        def serialize(writer: EoWriter, data: "QuestAcceptClientPacket.ReplyTypeDataOk") -> None:
            """
            Serializes an instance of `QuestAcceptClientPacket.ReplyTypeDataOk` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (QuestAcceptClientPacket.ReplyTypeDataOk): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                writer.add_char(0)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "QuestAcceptClientPacket.ReplyTypeDataOk":
            """
            Deserializes an instance of `QuestAcceptClientPacket.ReplyTypeDataOk` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                QuestAcceptClientPacket.ReplyTypeDataOk: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                reader.get_char()
                result = QuestAcceptClientPacket.ReplyTypeDataOk()
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"QuestAcceptClientPacket.ReplyTypeDataOk(byte_size={repr(self._byte_size)})"

    class ReplyTypeDataLink:
        """
        Data associated with reply_type value DialogReply.Link
        """
        _byte_size: int = 0
        _action: int

        def __init__(self, *, action: int):
            """
            Create a new instance of QuestAcceptClientPacket.ReplyTypeDataLink.

            Args:
                action (int): (Value range is 0-252.)
            """
            self._action = action

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def action(self) -> int:
            return self._action

        @staticmethod
        def serialize(writer: EoWriter, data: "QuestAcceptClientPacket.ReplyTypeDataLink") -> None:
            """
            Serializes an instance of `QuestAcceptClientPacket.ReplyTypeDataLink` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (QuestAcceptClientPacket.ReplyTypeDataLink): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._action is None:
                    raise SerializationError("action must be provided.")
                writer.add_char(data._action)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "QuestAcceptClientPacket.ReplyTypeDataLink":
            """
            Deserializes an instance of `QuestAcceptClientPacket.ReplyTypeDataLink` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                QuestAcceptClientPacket.ReplyTypeDataLink: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                action = reader.get_char()
                result = QuestAcceptClientPacket.ReplyTypeDataLink(action=action)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"QuestAcceptClientPacket.ReplyTypeDataLink(byte_size={repr(self._byte_size)}, action={repr(self._action)})"



import os
from pathlib import Path
from xml.etree import ElementTree

from protocol_code_generator.generate.code_block import CodeBlock
from protocol_code_generator.generate.object_code_generator import ObjectCodeGenerator
from protocol_code_generator.generate.python_file import PythonFile
from protocol_code_generator.type.enum_type import EnumType
from protocol_code_generator.type.struct_type import StructType
from protocol_code_generator.type.type_factory import TypeFactory
from protocol_code_generator.util.docstring_utils import generate_docstring
from protocol_code_generator.util.name_utils import pascal_case_to_snake_case
from protocol_code_generator.util.xml_utils import (
    get_comment,
    get_instructions,
    get_required_string_attribute,
)


class ProtocolFile:
    def __init__(self, path, protocol):
        self.path = path
        self.protocol = protocol


class ProtocolCodeGenerator:
    def __init__(self, input_root):
        self._input_root = input_root.as_posix()
        self._output_root = None
        self._protocol_files = []
        self._exports = []
        self._packet_paths = {}
        self._type_factory = TypeFactory()

    def generate(self, output_root):
        self._output_root = output_root.as_posix()
        try:
            self._index_protocol_files()
            self._generate_source_files()
        finally:
            self._protocol_files.clear()
            self._exports.clear()
            self._packet_paths.clear()
            self._type_factory.clear()

    def _index_protocol_files(self):
        for root, _, files in os.walk(self._input_root):
            if "protocol.xml" in files:
                protocol_file = Path(os.path.join(root, "protocol.xml")).as_posix()
                self._index_protocol_file(protocol_file)

    def _index_protocol_file(self, path):
        try:
            tree = ElementTree.parse(path)
            protocol = tree.getroot()

            if protocol.tag != 'protocol':
                raise RuntimeError("Expected a root <protocol> element.")

            self._protocol_files.append(ProtocolFile(path, protocol))

            enum_elements = protocol.findall("enum")
            struct_elements = protocol.findall("struct")
            packet_elements = protocol.findall("packet")

            source_path = os.path.dirname(os.path.relpath(path, self._input_root))
            source_path = Path(source_path).as_posix()

            for protocol_enum in enum_elements:
                if not self._type_factory.define_custom_type(protocol_enum, source_path):
                    raise RuntimeError(
                        get_required_string_attribute(protocol_enum, 'name')
                        + " type cannot be redefined."
                    )

            for protocol_struct in struct_elements:
                if not self._type_factory.define_custom_type(protocol_struct, source_path):
                    raise RuntimeError(
                        get_required_string_attribute(protocol_struct, 'name')
                        + " type cannot be redefined."
                    )

            declared_packets = set()
            for protocol_packet in packet_elements:
                packet_identifier = (
                    get_required_string_attribute(protocol_packet, 'family')
                    + '_'
                    + get_required_string_attribute(protocol_packet, 'action')
                )
                if packet_identifier in declared_packets:
                    raise RuntimeError(
                        f"{packet_identifier} packet cannot be redefined in the same file."
                    )
                declared_packets.add(packet_identifier)
                self._packet_paths[protocol_packet] = source_path

        except Exception as e:
            print(f"Failed to read {path}: {e}")
            raise e

    def _generate_source_files(self):
        for protocol_file in self._protocol_files:
            self._generate_source_file(protocol_file)

    def _generate_source_file(self, protocol_file):
        protocol = protocol_file.protocol
        python_files = [
            *map(self._generate_enum, protocol.findall("enum")),
            *map(self._generate_struct, protocol.findall("struct")),
            *map(self._generate_packet, protocol.findall("packet")),
        ]

        generated_init = CodeBlock()

        for python_file in python_files:
            python_file.write(self._output_root)

            absolute_package_path = Path(python_file.relative_path).with_suffix('')
            absolute_package_path = os.path.join("eolib/protocol/_generated", absolute_package_path)
            absolute_package_path = '.'.join(Path(absolute_package_path).parts)

            generated_init.add_import("*", absolute_package_path)

        relative_path = Path(os.path.relpath(protocol_file.path, self._input_root)).as_posix()
        path = os.path.join(os.path.dirname(relative_path), "__init__.py")
        path = Path(path).as_posix()

        eo_protocol_url = "https://github.com/cirras/eo-protocol/tree/master/xml/" + relative_path

        public_package = os.path.join("eolib/protocol", relative_path)
        public_package = os.path.dirname(public_package)
        public_package = '.'.join(Path(public_package).parts)

        docstring = (
            CodeBlock()
            .add_line('"""')
            .add_line(
                'Data structures generated from the '
                + f'[eo-protocol]({eo_protocol_url}){{target="_blank"}} XML specification.'
            )
            .add_line()
            .add_line('Warning:')
            .add_line('  - This subpackage should not be directly imported. ')
            .add_line(f'  - Instead, import [{public_package}][] (or the top-level `eolib`).')
            .add_line('"""')
        )

        generated_init_file = PythonFile(path, generated_init, module_docstring=docstring)
        generated_init_file.write(self._output_root)

    def _generate_enum(self, protocol_enum):
        type_name = get_required_string_attribute(protocol_enum, "name")
        type_ = self._type_factory.get_type(type_name)
        if not isinstance(type_, EnumType):
            raise RuntimeError(f"{type_name} is not a valid EnumType.")

        print(f"Generating enum: {type_.name}")

        code_block = CodeBlock()
        code_block.add_line(f"class {type_name}(IntEnum, metaclass=ProtocolEnumMeta):")
        code_block.indent()
        code_block.add_code_block(generate_docstring(get_comment(protocol_enum)))

        for protocol_value in protocol_enum.findall("value"):
            value_name = get_required_string_attribute(protocol_value, "name")
            value = type_.get_enum_value_by_name(value_name)
            code_block.add_line(f"{value.python_name} = {value.ordinal_value}")
            code_block.add_code_block(generate_docstring(get_comment(protocol_value)))

        code_block.unindent()
        code_block.add_import("IntEnum", "enum")
        code_block.add_import("ProtocolEnumMeta", "eolib.protocol.protocol_enum_meta")

        relative_path = os.path.join(type_.source_path, pascal_case_to_snake_case(type_name))
        self._exports.append(relative_path)

        return PythonFile(relative_path + ".py", code_block)

    def _generate_struct(self, protocol_struct):
        type_name = get_required_string_attribute(protocol_struct, "name")
        type_ = self._type_factory.get_type(type_name)
        if not isinstance(type_, StructType):
            raise RuntimeError(f"{type_name} is not a valid StructType.")

        print(f"Generating struct: {type_.name}")

        object_code_generator = ObjectCodeGenerator(type_.name, self._type_factory)
        for instruction in get_instructions(protocol_struct):
            object_code_generator.generate_instruction(instruction)

        relative_path = os.path.join(type_.source_path, pascal_case_to_snake_case(type_name))
        self._exports.append(relative_path)

        object_code_generator.data.docstring.add_code_block(
            generate_docstring(get_comment(protocol_struct))
        )

        code_block = object_code_generator.code

        return PythonFile(relative_path + ".py", code_block)

    def _generate_packet(self, protocol_packet):
        source_path = self._packet_paths.get(protocol_packet)
        packet_suffix = self._make_packet_suffix(source_path)
        family_attribute = get_required_string_attribute(protocol_packet, "family")
        action_attribute = get_required_string_attribute(protocol_packet, "action")
        packet_type_name = family_attribute + action_attribute + packet_suffix

        print(f"Generating packet: {packet_type_name}")

        family_type = self._type_factory.get_type("PacketFamily")
        if not isinstance(family_type, EnumType):
            raise RuntimeError("PacketFamily enum is missing.")

        action_type = self._type_factory.get_type("PacketAction")
        if not isinstance(action_type, EnumType):
            raise RuntimeError("PacketAction enum is missing.")

        family_enum_value = family_type.get_enum_value_by_name(family_attribute)
        if not family_enum_value:
            raise RuntimeError(f'Unknown packet family "{family_attribute}"')

        action_enum_value = action_type.get_enum_value_by_name(action_attribute)
        if not action_enum_value:
            raise RuntimeError(f'Unknown packet action "{action_attribute}"')

        object_code_generator = ObjectCodeGenerator(packet_type_name, self._type_factory)
        for instruction in get_instructions(protocol_packet):
            object_code_generator.generate_instruction(instruction)

        data = object_code_generator.data
        data.super_interfaces.append("Packet")

        data.add_method(
            CodeBlock()
            .add_line("@staticmethod")
            .add_line("def family() -> PacketFamily:")
            .indent()
            .add_line('"""')
            .add_line("Returns the packet family associated with this packet.")
            .add_line()
            .add_line("Returns:")
            .add_line("    PacketFamily: The packet family associated with this packet.")
            .add_line('"""')
            .add_line(f"return PacketFamily.{family_enum_value.python_name}")
            .unindent()
        )

        data.add_method(
            CodeBlock()
            .add_line("@staticmethod")
            .add_line("def action() -> PacketAction:")
            .indent()
            .add_line('"""')
            .add_line("Returns the packet action associated with this packet.")
            .add_line()
            .add_line("Returns:")
            .add_line("    PacketAction: The packet action associated with this packet.")
            .add_line('"""')
            .add_line(f"return PacketAction.{action_enum_value.python_name}")
            .unindent()
        )

        data.add_method(
            CodeBlock()
            .add_line("def write(self, writer):")
            .indent()
            .add_line('"""')
            .add_line("Serializes and writes this packet to the provided EoWriter.")
            .add_line()
            .add_line("Args:")
            .add_line("    writer (EoWriter): the writer that this packet will be written to.")
            .add_line('"""')
            .add_line(f"{packet_type_name}.serialize(writer, self)")
            .unindent()
        )

        object_code_generator.data.docstring.add_code_block(
            generate_docstring(get_comment(protocol_packet))
        )

        code_block = object_code_generator.code
        code_block.add_import("Packet", "eolib.protocol.net.packet")
        code_block.add_import("PacketFamily", "eolib.protocol._generated.net.packet_family")
        code_block.add_import("PacketAction", "eolib.protocol._generated.net.packet_action")

        relative_path = os.path.join(source_path, pascal_case_to_snake_case(packet_type_name))
        self._exports.append(relative_path)

        return PythonFile(relative_path + ".py", code_block)

    @staticmethod
    def _make_packet_suffix(path):
        if path == "net/client":
            return "ClientPacket"
        elif path == "net/server":
            return "ServerPacket"
        else:
            raise ValueError(f"Cannot create packet name suffix for path {path}")



from typing import Optional
from eolib.data.number_encoding_utils import decode_number
from eolib.data.string_encoding_utils import decode_string


class EoReader(object):
    """
    A class for reading EO data from a sequence of bytes.

    `EoReader` features a chunked reading mode, which is important for accurate emulation of
    the official game client.

    See documentation for chunked reading:
    https://github.com/Cirras/eo-protocol/blob/master/docs/chunks.md
    """

    _data: memoryview
    _position: int
    _chunked_reading_mode: bool
    _chunk_start: int
    _next_break: int

    def __init__(self, data: bytes):
        """
        Creates a new `EoReader` instance for the specified data.

        Args:
            data (bytes): The byte array containing the input data.
        """
        self._data = memoryview(data)
        self._position = 0
        self._chunked_reading_mode = False
        self._chunk_start = 0
        self._next_break = -1

    def slice(self, index: Optional[int] = None, length: Optional[int] = None) -> "EoReader":
        """
        Creates a new `EoReader` whose input data is a shared subsequence of this reader's
        data.

        The input data of the new reader will start at position `index` in this reader and contain
        up to `length` bytes. The two reader's position and chunked reading mode will be
        independent.

        The new reader's position will be zero, and its chunked reading mode will be false.

        Args:
            index (int, optional): The position in this reader at which the data of the new reader
                will start; must be non-negative. Defaults to the current reader position.
            length (int, optional): The length of the shared subsequence of data to supply to the
                new reader; must be non-negative. Defaults to the length of the remaining data
                starting from `index`.

        Returns:
            EoReader: The new reader.

        Raises:
            ValueError: If `index` or `length` is negative.
        """
        if index is None:
            index = self.position
        if length is None:
            length = max(0, len(self._data) - index)

        if index < 0:
            raise ValueError(f"negative index: {index}")

        if length < 0:
            raise ValueError(f"negative length: {length}")

        begin = max(0, min(len(self._data), index))
        end = begin + min(len(self._data) - begin, length)

        return EoReader(self._data[begin:end])

    def get_byte(self) -> int:
        """
        Reads a raw byte from the input data.

        Returns:
            int: A raw byte.
        """
        return self._read_byte()

    def get_bytes(self, length: int) -> bytearray:
        """
        Reads an array of raw bytes from the input data.

        Args:
            length (int): The number of bytes to read.

        Returns:
            bytearray: An array of raw bytes.
        """
        return self._read_bytes(length)

    def get_char(self) -> int:
        """
        Reads an encoded 1-byte integer from the input data.

        Returns:
            int: A decoded 1-byte integer.
        """
        return decode_number(self._read_bytes(1))

    def get_short(self) -> int:
        """
        Reads an encoded 2-byte integer from the input data.

        Returns:
            int: A decoded 2-byte integer.
        """
        return decode_number(self._read_bytes(2))

    def get_three(self) -> int:
        """
        Reads an encoded 3-byte integer from the input data.

        Returns:
            int: A decoded 3-byte integer.
        """
        return decode_number(self._read_bytes(3))

    def get_int(self) -> int:
        """
        Reads an encoded 4-byte integer from the input data.

        Returns:
            int: A decoded 4-byte integer.
        """
        return decode_number(self._read_bytes(4))

    def get_string(self) -> str:
        """
        Reads a string from the input data.

        Returns:
            str: A string.
        """
        string_bytes = self._read_bytes(self.remaining)
        return self._decode_ansi(string_bytes)

    def get_fixed_string(self, length: int, padded: bool = False) -> str:
        """
        Reads a string with a fixed length from the input data.

        Args:
            length (int): The length of the string.
            padded (bool, optional): True if the string is padded with trailing `0xFF` bytes.

        Returns:
            str: A decoded string.

        Raises:
            ValueError: If the length is negative.
        """
        if length < 0:
            raise ValueError("Negative length")
        bytes_ = self._read_bytes(length)
        if padded:
            bytes_ = self._remove_padding(bytes_)
        return self._decode_ansi(bytes_)

    def get_encoded_string(self) -> str:
        """
        Reads an encoded string from the input data.

        Returns:
            str: A decoded string.
        """
        bytes_ = self._read_bytes(self.remaining)
        decode_string(bytes_)
        return self._decode_ansi(bytes_)

    def get_fixed_encoded_string(self, length: int, padded: bool = False) -> str:
        """
        Reads an encoded string with a fixed length from the input data.

        Args:
            length (int): The length of the string.
            padded (bool, optional): True if the string is padded with trailing `0xFF` bytes.

        Returns:
            str: A decoded string.

        Raises:
            ValueError: If the length is negative.
        """
        if length < 0:
            raise ValueError("Negative length")
        bytes_ = self._read_bytes(length)
        decode_string(bytes_)
        if padded:
            bytes_ = self._remove_padding(bytes_)
        return self._decode_ansi(bytes_)

    @property
    def chunked_reading_mode(self) -> bool:
        """
        bool: Gets or sets the chunked reading mode for the reader.

        In chunked reading mode:
        - The reader will treat `0xFF` bytes as the end of the current chunk.
        - `next_chunk()` can be called to move to the next chunk.
        """
        return self._chunked_reading_mode

    @chunked_reading_mode.setter
    def chunked_reading_mode(self, chunked_reading_mode: bool) -> None:
        self._chunked_reading_mode = chunked_reading_mode
        if self._next_break == -1:
            self._next_break = self._find_next_break_index()

    @property
    def remaining(self) -> int:
        """
        int: If chunked reading mode is enabled, gets the number of bytes remaining in the current
            chunk. Otherwise, gets the total number of bytes remaining in the input data.
        """
        if self.chunked_reading_mode:
            return self._next_break - min(self.position, self._next_break)
        else:
            return len(self._data) - self.position

    def next_chunk(self) -> None:
        """
        Moves the reader position to the start of the next chunk in the input data.

        Raises:
            RuntimeError: If not in chunked reading mode.
        """
        if not self.chunked_reading_mode:
            raise RuntimeError("Not in chunked reading mode.")

        self._position = self._next_break
        if self._position < len(self._data):
            # Skip the break byte
            self._position += 1

        self._chunk_start = self._position
        self._next_break = self._find_next_break_index()

    @property
    def position(self) -> int:
        """
        int: Gets the current position in the input data.
        """
        return self._position

    def _read_byte(self) -> int:
        """
        Reads a raw byte from the input data.

        Returns:
            int: A raw byte.
        """
        if self.remaining > 0:
            byte = self._data[self._position]
            self._position += 1
            return byte
        return 0

    def _read_bytes(self, length) -> bytearray:
        """
        Reads an array of raw bytes from the input data.

        Args:
            length (int): The number of bytes to read.

        Returns:
            bytearray: An array of raw bytes.
        """
        length = min(length, self.remaining)

        result = bytearray(self._data[self._position : self._position + length])
        self._position += length

        return result

    def _find_next_break_index(self) -> int:
        """
        Finds the index of the next break byte (0xFF) in the input data.

        Returns:
            int: The index of the next break byte, or the length of the data if not found.
        """
        for i in range(self._chunk_start, len(self._data)):
            if self._data[i] == 0xFF:
                return i
        return len(self._data)

    @staticmethod
    def _remove_padding(array: bytearray) -> bytearray:
        """
        Removes padding (trailing 0xFF bytes) from a sequence of bytes.

        Args:
            array (bytearray): The sequence of bytes.

        Returns:
            bytearray: The bytes without padding.
        """
        padding_start = array.find(bytes([0xFF]))
        if padding_start != -1:
            return array[:padding_start]
        return array

    @staticmethod
    def _decode_ansi(bytes: bytearray) -> str:
        """
        Decodes windows-1252 bytes to a string.

        Args:
            bytes (bytearray): The sequence of bytes to decode.

        Returns:
            str: The decoded string.
        """
        return bytes.decode('windows-1252', 'replace')



"""
EO protocol data structures.

See Also:
  - [eolib.protocol._generated][]
"""

from .serialization_error import *

from .map import *
from .net import *
from .pub import *

from ._generated import *



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ...serialization_error import SerializationError
from ....data.eo_writer import EoWriter
from ....data.eo_reader import EoReader

class Version:
    """
    Client version
    """
    _byte_size: int = 0
    _major: int
    _minor: int
    _patch: int

    def __init__(self, *, major: int, minor: int, patch: int):
        """
        Create a new instance of Version.

        Args:
            major (int): (Value range is 0-252.)
            minor (int): (Value range is 0-252.)
            patch (int): (Value range is 0-252.)
        """
        self._major = major
        self._minor = minor
        self._patch = patch

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def major(self) -> int:
        return self._major

    @property
    def minor(self) -> int:
        return self._minor

    @property
    def patch(self) -> int:
        return self._patch

    @staticmethod
    def serialize(writer: EoWriter, data: "Version") -> None:
        """
        Serializes an instance of `Version` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (Version): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._major is None:
                raise SerializationError("major must be provided.")
            writer.add_char(data._major)
            if data._minor is None:
                raise SerializationError("minor must be provided.")
            writer.add_char(data._minor)
            if data._patch is None:
                raise SerializationError("patch must be provided.")
            writer.add_char(data._patch)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "Version":
        """
        Deserializes an instance of `Version` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            Version: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            major = reader.get_char()
            minor = reader.get_char()
            patch = reader.get_char()
            result = Version(major=major, minor=minor, patch=patch)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"Version(byte_size={repr(self._byte_size)}, major={repr(self._major)}, minor={repr(self._minor)}, patch={repr(self._patch)})"
